import math
from collections.abc import Iterator, Mapping
from typing import Any

from phylogenie.core.node import Node
from phylogenie.mixins import MetadataMixin


class Tree(MetadataMixin):
    def __init__(self, root: Node):
        super().__init__()
        self.root = root

    # ----------------
    # Basic properties
    # ----------------
    # Properties related to parent-child relationships.

    def add_subtree(self, parent: Node, subtree: "Tree") -> "Tree":
        parent.add_child(subtree.root)
        return self

    def get_leaves(self) -> tuple[Node, ...]:
        return tuple(node for node in self if node.is_leaf())

    def get_internal_nodes(self) -> tuple[Node, ...]:
        return tuple(node for node in self if node.is_internal())

    def is_binary(self) -> bool:
        return all(len(node.children) in (0, 2) for node in self)

    # --------------
    # Tree traversal
    # --------------
    # Methods for traversing the tree in various orders.

    def preorder_traversal(self) -> Iterator[Node]:
        yield from self.root.iter_preorder()

    def inorder_traversal(self) -> Iterator[Node]:
        yield from self.root.iter_inorder()

    def postorder_traversal(self) -> Iterator[Node]:
        yield from self.root.iter_postorder()

    def breadth_first_traversal(self) -> Iterator[Node]:
        yield from self.root.iter_breadth_first()

    # ---------------
    # Tree properties
    # ---------------
    # Properties and methods related to tree metrics like leaf count, depth, height, etc.

    @property
    def n_leaves(self) -> int:
        return len(self.get_leaves())

    @property
    def leaf_counts(self) -> dict[Node, int]:
        n_leaves: dict[Node, int] = {}
        for node in self.postorder_traversal():
            n_leaves[node] = sum(n_leaves[child] for child in node.children) or 1
        return n_leaves

    @property
    def height_level(self) -> int:
        return self.height_levels[self.root]

    @property
    def height_levels(self) -> dict[Node, int]:
        height_levels: dict[Node, int] = {}
        for node in self.postorder_traversal():
            height_levels[node] = (
                0
                if node.is_leaf()
                else max(1 + height_levels[child] for child in node.children)
            )
        return height_levels

    @property
    def height(self) -> float:
        return self.heights[self.root]

    @property
    def heights(self) -> dict[Node, float]:
        heights: dict[Node, float] = {}
        for node in self.postorder_traversal():
            heights[node] = (
                0
                if node.is_leaf()
                else max(
                    child.branch_length_or_raise() + heights[child]
                    for child in node.children
                )
            )
        return heights

    @property
    def depth_level(self) -> float:
        return self.height_level

    @property
    def depth_levels(self) -> dict[Node, int]:
        depth_levels: dict[Node, int] = {self.root: 0}
        for node in self.root.iter_descendants():
            depth_levels[node] = depth_levels[node.parent] + 1  # pyright: ignore
        return depth_levels

    @property
    def depth(self) -> float:
        return self.height

    @property
    def depths(self) -> dict[Node, float]:
        depths: dict[Node, float] = {self.root: 0.0}
        for node in self.root.iter_descendants():
            parent_depth = depths[node.parent]  # pyright: ignore
            depths[node] = node.branch_length_or_raise() + parent_depth
        return depths

    @property
    def times(self) -> dict[Node, float]:
        root_time = self.root.branch_length_or_raise()
        return {node: root_time + depth for node, depth in self.depths.items()}

    @property
    def age(self) -> float:
        return self.height + self.root.branch_length_or_raise()

    @property
    def ages(self) -> dict[Node, float]:
        ages: dict[Node, float] = {self.root: self.height}
        for node in self.root.iter_descendants():
            ages[node] = ages[node.parent] - node.branch_length_or_raise()  # pyright: ignore
        return ages

    def compute_sackin_index(self, normalize: bool = False) -> float:
        sackin_index = sum(
            dl for node, dl in self.depth_levels.items() if node.is_leaf()
        )
        if normalize:
            if not self.is_binary():
                raise ValueError(
                    "Normalized Sackin index is only defined for binary trees."
                )
            n = self.n_leaves
            h = math.floor(math.log2(n))
            min_sackin_index = n * (h + 2) - 2 ** (h + 1)
            max_sackin_index = n * (n - 1) / 2
            return (sackin_index - min_sackin_index) / (
                max_sackin_index - min_sackin_index
            )
        return sackin_index

    def compute_mean_leaf_pairwise_distance(self) -> float:
        leaves = self.get_leaves()
        n_leaves = len(leaves)
        if n_leaves < 2:
            return 0.0

        total_distance = sum(
            leaves[i].get_distance(leaves[j])
            for i in range(n_leaves)
            for j in range(i + 1, n_leaves)
        )
        return total_distance / math.comb(n_leaves, 2)

    # -------------
    # Miscellaneous
    # -------------
    # Other useful miscellaneous methods.

    def ladderize(self, key: Mapping[Node, Any] | None = None) -> None:
        if key is None:
            key = self.leaf_counts

        for node in self:
            node.sort_children(key=lambda child: key[child])

    def get_node(self, name: str) -> Node:
        for node in self:
            if node.name == name:
                return node
        raise ValueError(f"Node {name} not found.")

    def copy(self):
        new_tree = Tree(self.root.copy())
        new_tree.update(self.metadata)
        return new_tree

    # ----------------
    # Dunder methods
    # ----------------
    # Special methods for standard behaviors like iteration, length, and representation.

    def __iter__(self) -> Iterator[Node]:
        return self.preorder_traversal()

    def __len__(self) -> int:
        return sum(1 for _ in self)

    def __repr__(self) -> str:
        return f"TreeNode(metadata={self.metadata})"



from phylogenie.treesimulator.events.base import Event
from phylogenie.treesimulator.events.contact_tracing import (
    BirthWithContactTracing,
    SamplingWithContactTracing,
    get_contact_tracing_events,
)
from phylogenie.treesimulator.events.core import (
    Birth,
    Death,
    Migration,
    Sampling,
    get_BD_events,
    get_BDEI_events,
    get_BDSS_events,
    get_canonical_events,
    get_epidemiological_events,
    get_FBD_events,
)

__all__ = [
    "Birth",
    "BirthWithContactTracing",
    "Death",
    "Event",
    "Migration",
    "Sampling",
    "SamplingWithContactTracing",
    "get_BD_events",
    "get_BDEI_events",
    "get_BDSS_events",
    "get_canonical_events",
    "get_contact_tracing_events",
    "get_epidemiological_events",
    "get_FBD_events",
]



from bisect import bisect_right
from collections.abc import Callable
from typing import Any, TypeGuard, Union

import phylogenie.typeguards as tg
import phylogenie.typings as pgt

SkylineParameterLike = Union[pgt.Scalar, "SkylineParameter"]


def is_skyline_parameter_like(x: Any) -> TypeGuard[SkylineParameterLike]:
    return isinstance(x, pgt.Scalar | SkylineParameter)


def is_many_skyline_parameters_like(
    x: Any,
) -> TypeGuard[pgt.Many[SkylineParameterLike]]:
    return tg.is_many(x) and all(is_skyline_parameter_like(v) for v in x)


class SkylineParameter:
    def __init__(
        self,
        value: pgt.OneOrManyScalars,
        change_times: pgt.ManyScalars | None = None,
    ):
        if isinstance(value, pgt.Scalar):
            value = [value]
        elif not tg.is_many_scalars(value):
            raise TypeError(
                f"It is impossible to create a SkylineParameter from `value` {value} of type {type(value)}. Please provide a scalar or a sequence of scalars."
            )

        if change_times is None:
            change_times = []
        elif not tg.is_many_scalars(change_times):
            raise TypeError(
                f"It is impossible to create a SkylineParameter from `change_times` {change_times} of type {type(change_times)}. Please provide a sequence of scalars."
            )

        if len(value) != len(change_times) + 1:
            raise ValueError(
                f"`value` must have exactly one more element than `change_times` (got value={value} of length {len(value)} and change_times={change_times} of length {len(change_times)})."
            )
        if any(t1 >= t2 for t1, t2 in zip(change_times, change_times[1:])):
            raise ValueError(
                f"`change_times` must be sorted in strictly increasing order "
                f"(got change_times={change_times})."
            )
        if any(t < 0 for t in change_times):
            raise ValueError(
                f"`change_times` must be non-negative (got change_times={change_times})."
            )

        self._value = [value[0]]
        self._change_times: list[pgt.Scalar] = []
        for i in range(1, len(value)):
            if value[i] != value[i - 1]:
                self._value.append(value[i])
                self._change_times.append(change_times[i - 1])

    @property
    def value(self) -> pgt.Vector1D:
        return tuple(self._value)

    @property
    def change_times(self) -> pgt.Vector1D:
        return tuple(self._change_times)

    def get_value_at_time(self, t: pgt.Scalar) -> pgt.Scalar:
        if t < 0:
            raise ValueError(f"Time cannot be negative (got t={t}).")
        return self.value[bisect_right(self.change_times, t)]

    def _operate(
        self,
        other: SkylineParameterLike,
        f: Callable[[pgt.Scalar, pgt.Scalar], pgt.Scalar],
    ) -> "SkylineParameter":
        if is_skyline_parameter_like(other):
            other = skyline_parameter(other)
        else:
            return NotImplemented
        change_times = sorted(set(self.change_times + other.change_times))
        value = [
            f(self.get_value_at_time(t), other.get_value_at_time(t))
            for t in (0, *change_times)
        ]
        return SkylineParameter(value, change_times)

    def __add__(self, other: SkylineParameterLike) -> "SkylineParameter":
        return self._operate(other, lambda x, y: x + y)

    def __radd__(self, other: pgt.Scalar) -> "SkylineParameter":
        return self._operate(other, lambda x, y: y + x)

    def __sub__(self, other: SkylineParameterLike) -> "SkylineParameter":
        return self._operate(other, lambda x, y: x - y)

    def __rsub__(self, other: pgt.Scalar) -> "SkylineParameter":
        return self._operate(other, lambda x, y: y - x)

    def __mul__(self, other: SkylineParameterLike) -> "SkylineParameter":
        return self._operate(other, lambda x, y: x * y)

    def __rmul__(self, other: pgt.Scalar) -> "SkylineParameter":
        return self._operate(other, lambda x, y: y * x)

    def __truediv__(self, other: SkylineParameterLike) -> "SkylineParameter":
        return self._operate(other, lambda x, y: x / y)

    def __rtruediv__(self, other: pgt.Scalar) -> "SkylineParameter":
        return self._operate(other, lambda x, y: y / x)

    def __bool__(self) -> bool:
        return any(self.value)

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, SkylineParameter) and (
            self.value == other.value and self.change_times == other.change_times
        )

    def __repr__(self) -> str:
        return f"SkylineParameter(value={list(self.value)}, change_times={list(self.change_times)})"


def skyline_parameter(x: SkylineParameterLike) -> SkylineParameter:
    return SkylineParameter(x) if isinstance(x, pgt.Scalar) else x



import time
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Any, Callable

import joblib
import numpy as np
import pandas as pd
from numpy.random import default_rng
from tqdm import tqdm

from phylogenie.core import Tree
from phylogenie.io import dump_newick
from phylogenie.treesimulator.events import Event
from phylogenie.treesimulator.model import Model


def simulate_tree(
    events: Sequence[Event],
    n_tips: int | None = None,
    max_time: float = np.inf,
    init_state: str | None = None,
    sampling_probability_at_present: float = 0.0,
    seed: int | None = None,
    timeout: float = np.inf,
    acceptance_criterion: Callable[[Tree], bool] | None = None,
    logs: dict[str, Callable[[Tree], Any]] | None = None,
) -> tuple[Tree, dict[str, Any]]:
    if (max_time != np.inf) == (n_tips is not None):
        raise ValueError("Exactly one of max_time or n_tips must be specified.")
    if sampling_probability_at_present and max_time == np.inf:
        raise ValueError(
            "sampling_probability_at_present can only be used with max_time."
        )

    states = {e.state for e in events if e.state}
    if init_state is None and len(states) > 1:
        raise ValueError(
            "Init state must be provided for models with more than one state."
        )
    elif init_state is None:
        (init_state,) = states
    elif init_state not in states:
        raise ValueError(f"Init state {init_state} not found in event states: {states}")

    rng = default_rng(seed)
    start_clock = time.perf_counter()
    while True:
        model = Model(init_state)
        metadata: dict[str, Any] = {}
        current_time = 0.0
        change_times = sorted(set(t for e in events for t in e.rate.change_times))
        next_change_time = change_times.pop(0) if change_times else np.inf

        while current_time < max_time and (n_tips is None or model.n_sampled < n_tips):
            if time.perf_counter() - start_clock > timeout:
                raise TimeoutError("Simulation timed out.")

            propensities = [e.get_propensity(model, current_time) for e in events]
            if not any(propensities):
                break

            time_step = rng.exponential(1 / sum(propensities))
            if current_time + time_step >= next_change_time:
                current_time = next_change_time
                next_change_time = change_times.pop(0) if change_times else np.inf
                continue
            if current_time + time_step >= max_time:
                current_time = max_time
                break
            current_time += time_step

            event_idx = np.searchsorted(
                np.cumsum(propensities) / sum(propensities), rng.random()
            )
            event = events[int(event_idx)]
            event_metadata = event.apply(model, current_time, rng)
            if event_metadata is not None:
                metadata.update(event_metadata)

        if current_time != max_time and model.n_sampled != n_tips:
            continue

        for individual in model.get_population():
            if rng.random() < sampling_probability_at_present:
                model.sample(individual, current_time, True)

        tree = model.get_sampled_tree()

        if acceptance_criterion is not None and not acceptance_criterion(tree):
            continue

        if logs is not None:
            for key, func in logs.items():
                metadata[key] = func(tree)

        return (tree, metadata)


def generate_trees(
    output_dir: str | Path,
    n_trees: int,
    events: Sequence[Event],
    n_tips: int | None = None,
    max_time: float = np.inf,
    init_state: str | None = None,
    sampling_probability_at_present: float = 0.0,
    node_features: Mapping[str, str] | None = None,
    seed: int | None = None,
    n_jobs: int = -1,
    timeout: float = np.inf,
    acceptance_criterion: Callable[[Tree], bool] | None = None,
    logs: dict[str, Callable[[Tree], Any]] | None = None,
) -> pd.DataFrame:
    if isinstance(output_dir, str):
        output_dir = Path(output_dir)
    if output_dir.exists():
        raise FileExistsError(f"Output directory {output_dir} already exists")
    output_dir.mkdir(parents=True)

    def _simulate_tree(i: int, seed: int) -> dict[str, Any]:
        while True:
            try:
                tree, metadata = simulate_tree(
                    events=events,
                    n_tips=n_tips,
                    max_time=max_time,
                    init_state=init_state,
                    sampling_probability_at_present=sampling_probability_at_present,
                    seed=seed,
                    timeout=timeout,
                    acceptance_criterion=acceptance_criterion,
                    logs=logs,
                )
                metadata["file_id"] = i
                if node_features is not None:
                    for name, feature in node_features.items():
                        mapping = getattr(tree, feature)
                        for node in tree:
                            node[name] = mapping[node]
                dump_newick(tree, output_dir / f"{i}.nwk")
                return metadata
            except TimeoutError:
                print("Simulation timed out. Retrying with a different seed...")
            seed += 1

    rng = default_rng(seed)
    jobs = joblib.Parallel(n_jobs=n_jobs, return_as="generator_unordered")(
        joblib.delayed(_simulate_tree)(i=i, seed=int(rng.integers(2**32)))
        for i in range(n_trees)
    )

    return pd.DataFrame(
        [md for md in tqdm(jobs, f"Generating trees in {output_dir}...", n_trees)]
    )



from datetime import date
from pathlib import Path
from typing import Callable

from phylogenie.msa import MSA, Sequence


def load_fasta(
    fasta_file: str | Path,
    extract_time_from_id: Callable[[str], float | date] | None = None,
) -> MSA:
    sequences: list[Sequence] = []
    with open(fasta_file, "r") as f:
        for line in f:
            if not line.startswith(">"):
                raise ValueError(f"Invalid FASTA format: expected '>', got '{line[0]}'")
            id = line[1:].strip()
            time = None
            if extract_time_from_id is not None:
                time = extract_time_from_id(id)
            elif "|" in id:
                last_metadata = id.split("|")[-1]
                try:
                    time = float(last_metadata)
                except ValueError:
                    try:
                        time = date.fromisoformat(last_metadata)
                    except ValueError:
                        pass
            chars = next(f).strip()
            sequences.append(Sequence(id, chars, time))
    return MSA(sequences)


def dump_fasta(msa: MSA | list[Sequence], fasta_file: str | Path) -> None:
    with open(fasta_file, "w") as f:
        sequences = msa.sequences if isinstance(msa, MSA) else msa
        for seq in sequences:
            f.write(f">{seq.id}\n")
            f.write(f"{seq.chars}\n")



import os
from argparse import ArgumentParser
from glob import glob

from pydantic import TypeAdapter, ValidationError
from yaml import safe_load

from phylogenie.generators import DatasetGeneratorConfig
from phylogenie.generators.dataset import DatasetGenerator


def _format_validation_error(e: ValidationError) -> str:
    formatted_errors = [
        f"- {'.'.join(str(loc) for loc in err['loc'])}: {err['msg']} ({err['type']})"
        for err in e.errors()
    ]
    return "\n".join(formatted_errors)


def _generate_from_config_file(config_file: str):
    adapter: TypeAdapter[DatasetGenerator] = TypeAdapter(DatasetGeneratorConfig)
    with open(config_file, "r") as f:
        try:
            config = safe_load(f)
        except Exception as e:
            print(f"❌ Failed to parse {config_file}: {e}")
            exit(-1)
        try:
            generator = adapter.validate_python(config)
        except ValidationError as e:
            print("❌ Invalid configuration:")
            print(_format_validation_error(e))
            exit(-1)
    generator.generate()


def run(config_path: str) -> None:
    if os.path.isdir(config_path):
        for config_file in glob(os.path.join(config_path, "**/*.yaml"), recursive=True):
            _generate_from_config_file(config_file)
    else:
        _generate_from_config_file(config_path)


def main() -> None:
    parser = ArgumentParser(
        description="Generate dataset(s) starting from provided config(s)."
    )
    parser.add_argument(
        "config_path",
        type=str,
        help="Path to a config file or a directory containing config files.",
    )
    args = parser.parse_args()

    run(args.config_path)



from collections import deque
from collections.abc import Callable, Iterator
from typing import Any

from phylogenie.mixins import MetadataMixin


class Node(MetadataMixin):
    def __init__(self, name: str = "", branch_length: float | None = None):
        super().__init__()
        self.name = name
        self.branch_length = branch_length
        self._parent: Node | None = None
        self._children: list[Node] = []

    # ----------------
    # Basic properties
    # ----------------
    # Properties related to parent-child relationships.

    @property
    def children(self) -> tuple["Node", ...]:
        return tuple(self._children)

    @property
    def parent(self) -> "Node | None":
        return self._parent

    def add_child(self, child: "Node") -> "Node":
        if child.parent is not None:
            raise ValueError(f"Node {child.name} already has a parent.")
        child._parent = self
        self._children.append(child)
        return self

    def remove_child(self, child: "Node") -> None:
        self._children.remove(child)
        child._parent = None

    def set_parent(self, parent: "Node | None"):
        if self.parent is not None:
            self.parent.remove_child(self)
        self._parent = parent
        if parent is not None:
            parent._children.append(self)

    def is_leaf(self) -> bool:
        return not self.children

    def is_internal(self) -> bool:
        return not self.is_leaf()

    def branch_length_or_raise(self) -> float:
        if self.parent is None:
            return 0 if self.branch_length is None else self.branch_length
        if self.branch_length is None:
            raise ValueError(f"Branch length of node {self.name} is not set.")
        return self.branch_length

    # --------------
    # Tree traversal
    # --------------
    # Methods for traversing the tree in various orders.

    def iter_ancestors(self, stop: "Node | None" = None) -> Iterator["Node"]:
        node = self
        while True:
            if node.parent is None:
                if stop is None:
                    return
                raise ValueError("Reached root without encountering stop node.")
            node = node.parent
            if node == stop:
                return
            yield node

    def iter_upward(self, stop: "Node | None" = None) -> Iterator["Node"]:
        if self == stop:
            return
        yield self
        yield from self.iter_ancestors(stop=stop)

    def iter_descendants(self) -> Iterator["Node"]:
        for child in self.children:
            yield child
            yield from child.iter_descendants()

    def iter_preorder(self) -> Iterator["Node"]:
        yield self
        yield from self.iter_descendants()

    def iter_inorder(self) -> Iterator["Node"]:
        if self.is_leaf():
            yield self
            return
        if len(self.children) != 2:
            raise ValueError("Inorder traversal is only defined for binary trees.")
        left, right = self.children
        yield from left.iter_inorder()
        yield self
        yield from right.iter_inorder()

    def iter_postorder(self) -> Iterator["Node"]:
        for child in self.children:
            yield from child.iter_postorder()
        yield self

    def iter_breadth_first(self) -> Iterator["Node"]:
        queue: deque["Node"] = deque([self])
        while queue:
            node = queue.popleft()
            yield node
            queue.extend(node.children)

    # --------------
    # Paths and hops
    # --------------

    def get_mrca(self, node2: "Node") -> "Node":
        self_ancestors = set(self.iter_upward())
        for node2_ancestor in node2.iter_upward():
            if node2_ancestor in self_ancestors:
                return node2_ancestor
        raise ValueError(
            f"No common ancestor found between node {self} and node {node2}."
        )

    def get_path(self, node2: "Node") -> list["Node"]:
        mrca = self.get_mrca(node2)
        return [
            *self.iter_upward(stop=mrca.parent),
            *reversed(list(node2.iter_upward(stop=mrca))),
        ]

    def count_hops(self, node2: "Node") -> int:
        return len(self.get_path(node2)) - 1

    def get_distance(self, node2: "Node") -> float:
        mrca = self.get_mrca(node2)
        path = self.get_path(node2)
        path.remove(mrca)
        return sum(node.branch_length_or_raise() for node in path)

    # -------------
    # Miscellaneous
    # -------------
    # Miscellaneous methods like copying and representation.

    def copy(self):
        new_node = Node(self.name, self.branch_length)
        new_node.update(self.metadata)
        for child in self.children:
            new_node.add_child(child.copy())
        return new_node

    def sort_children(self, key: Callable[["Node"], Any]) -> None:
        self._children.sort(key=key)

    def __repr__(self) -> str:
        return f"Node(name='{self.name}', branch_length={self.branch_length}, metadata={self.metadata})"
